;; Generated package description from pabbrev.el  -*- no-byte-compile: t -*-
(define-package "pabbrev" "4.2.2.0.20230101.115226" "Predictive abbreviation expansion" 'nil :commit "6d9e97a6b6b7897e0fdd101299751219855114c0" :url "https://elpa.gnu.org/packages/pabbrev.html" :authors '(("Phillip Lord" . "phillip.lord@newcastle.ac.uk")) :maintainer '("Phillip Lord" . "phillip.lord@newcastle.ac.uk"))
